/*
 * \brief  Azure IoT Hub file upload helper class
 * \author Timo Nicolai
 * \date   2023-05-30
 */

/*
 *  Copyright (C) 2023 gapfruit AG
 */

#pragma once

/* Genode includes */
#include <base/log.h>
#include <util/reconstructible.h>
#include <util/string.h>

/* Azure includes */
extern "C" {

#include <azure/az_iot.h>

}

/* libc includes */
#include <dirent.h>
#include <errno.h>
#include <libgen.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>

/* libcurl includes */
extern "C" {

#include <curl/curl.h>

}

/* azure_iot_client includes */
#include <azure_iot_client/azure/types.h>
#include <azure_iot_client/http/status_codes.h>
#include <azure_iot_client/json/json.h>
#include <azure_iot_client/util/filesystem.h>
#include <azure_iot_client/util/timestamp.h>

namespace Azure_iot_client {
	class File_uploader;
}


class Azure_iot_client::File_uploader
{
	private:

		static constexpr char const *USER_AGENT_HEADER =
			"User-Agent: gapfruit_agent-azure_iot";

		Azure::Url _file_upload_endpoint;
		Azure::Url _file_upload_notify_endpoint;

		char const *_root_ca_certificates;
		char const *_device_certificate;
		char const *_device_private_key;

		JSON::Parser _json_parser { };
		JSON::Object _json_response { };

		static Path _get_cloud_file_name(Path const &file_path)
		{
			/* the filename of the blob we create in the cloud is the basename
			 * of the uploaded file plus a timestamp postfix to disambiguate
			 * subsequent uploads of the same file */

			/* copy file path into local buffer so we can pass it to basename */
			char file_path_buf[Path::capacity()];

			::strncpy(file_path_buf, file_path.string(), file_path.length());

			/* get local file name */
			Genode::Cstring local_file_name { ::basename(file_path_buf) };

			/* get cloud file name by appending timestamp */
			return Path {
				local_file_name, "_", timestamp(TIMESTAMP_FMT_UTC_COMPACT) };
		}

		/* see https://learn.microsoft.com/en-us/azure/storage/common/storage-sas-overview
		 * for information on storage account shared access signatures (SAS) */
		HTTP::Status_code _get_sas_uri_parameters(Path const &file_name) const
		{
			/* init request */
			auto curl { _curl_init_request(_file_upload_endpoint) };
			if (!curl)
				return HTTP::STATUS_INTERNAL_SERVER_ERROR;

			/* set request headers */
			::curl_slist *slist { nullptr };
			slist = ::curl_slist_append(slist, USER_AGENT_HEADER);
			slist = ::curl_slist_append(slist, "Content-Type: application/json");
			slist = ::curl_slist_append(slist, "Accept: application/json");
			::curl_easy_setopt(curl, CURLOPT_HTTPHEADER, slist);

			/* create POST data */
			using Postdata = Genode::String<1024 + 1>;
			Postdata postdata { "{\"blobName\":\"", file_name, "\"}" };
			::curl_easy_setopt(curl, CURLOPT_POSTFIELDS, postdata.string());

			::curl_easy_setopt(curl, CURLOPT_WRITEDATA, this);
			::curl_easy_setopt(curl,
			                   CURLOPT_WRITEFUNCTION,
			                   _curl_parse_json_response);

			/* send request */
			Genode::log("requesting SAS URI for uploading file '",
			            file_name, "' to blob storage");

			auto response_code { _curl_send_request(curl) };
			::curl_slist_free_all(slist);
			return response_code;
		}

		static HTTP::Status_code _get_sas_uri(
			JSON::Object const &sas_uri_parameters,
			Azure::Url &sas_uri)
		{
			try {
				auto host_name { sas_uri_parameters["hostName"] };
				auto container_name { sas_uri_parameters["containerName"] };
				auto blob_name { sas_uri_parameters["blobName"] };
				auto sas_token { sas_uri_parameters["sasToken"] };

				sas_uri = Azure::Url { "https://",
				                       host_name.as_string<char const *>(),
				                       "/",
				                       container_name.as_string<char const *>(),
				                       "/",
				                       blob_name.as_string<char const *>(),
				                       sas_token.as_string<char const *>() };

			} catch (JSON::Exception const &) {
				Genode::error("failed to construct SAS URI");
				return HTTP::STATUS_INTERNAL_SERVER_ERROR;
			}

			return HTTP::STATUS_OK;
		}

		HTTP::Status_code _upload_file(Path const &file_path,
		                               Azure::Url const &sas_uri) const
		{
			/* open file */
			Genode::Constructible<Readonly_file> file;
			::curl_off_t file_size { 0 };

			try {
				file.construct(file_path.string());
				file_size = static_cast<::curl_off_t>(file->size());

			} catch (Readonly_file::Open_exception const &) {
				Genode::error("failed to open file '", file_path, "' for transfer");
				return HTTP::STATUS_INTERNAL_SERVER_ERROR;

			} catch (Readonly_file::Read_exception const &) {
				Genode::error("failed to determine size of file '", file_path, "'");
				return HTTP::STATUS_INTERNAL_SERVER_ERROR;
			}

			/* init request */
			auto curl { _curl_init_request(sas_uri) };
			if (!curl)
				return HTTP::STATUS_INTERNAL_SERVER_ERROR;

			/* add request headers */
			::curl_slist *slist { nullptr };
			slist = ::curl_slist_append(slist, USER_AGENT_HEADER);
			slist = ::curl_slist_append(slist, "x-ms-blob-type: BlockBlob");
			::curl_easy_setopt(curl, CURLOPT_HTTPHEADER, slist);

			/* set file to upload */
			::curl_easy_setopt(curl, CURLOPT_UPLOAD, 1L);

			::curl_easy_setopt(curl, CURLOPT_READFUNCTION, _curl_read_file);
			::curl_easy_setopt(curl, CURLOPT_READDATA, &*file);
			::curl_easy_setopt(curl, CURLOPT_INFILESIZE_LARGE, file_size);

			::curl_easy_setopt(curl,
			                   CURLOPT_WRITEFUNCTION,
			                   _curl_ignore_response);

			/* send request */
			Genode::log("uploading file at '", file_path, "' to blob storage");

			auto response_code { _curl_send_request(curl) };
			::curl_slist_free_all(slist);
			return response_code;
		}

		HTTP::Status_code _notify_iot_hub(
			JSON::Object const &sas_uri_parameters,
			HTTP::Status_code upload_response_code) const
		{
			/* init request */
			auto curl { _curl_init_request(
				_file_upload_notify_endpoint.string()) };
			if (!curl)
				return HTTP::STATUS_INTERNAL_SERVER_ERROR;

			/* add request headers */
			curl_slist *slist { nullptr };
			slist = ::curl_slist_append(slist, USER_AGENT_HEADER);
			slist = ::curl_slist_append(slist, "Content-Type: application/json");
			::curl_easy_setopt(curl, CURLOPT_HTTPHEADER, slist);

			/* construct payload */
			JSON::Object payload;

			payload.add_key("correlationId",
			                sas_uri_parameters["correlationId"]);

			bool success { HTTP::okay(upload_response_code) };

			char const *description = success ? "File uploaded successfully"
			                                  : "File upload failed";

			payload.add_key("isSuccess",
			                JSON::Object::from_bool(success));
			payload.add_key("statusCode",
			                JSON::Object::from_int(upload_response_code));
			payload.add_key("statusDescription",
			                JSON::Object::from_string(description));

			/* append payload to POST data */
			char const *postdata { payload.repr() };
			::curl_easy_setopt(curl, CURLOPT_POSTFIELDS, postdata);

			/* send request */
			Genode::log("notifying IoT Hub of file upload, response code was ",
			            upload_response_code);

			auto response_code { _curl_send_request(curl) };
			::curl_slist_free_all(slist);
			return response_code;
		}

		CURL *_curl_init_request(Azure::Url const &url) const
		{
			auto curl { ::curl_easy_init() };
			if (!curl) {
				Genode::error("failed to initialize curl");
				return nullptr;
			}

			::curl_easy_setopt(curl, CURLOPT_URL, url.string());

			/* causes `curl_easy_perform` to fail on HTTP response codes >= 400 */
			::curl_easy_setopt(curl, CURLOPT_FAILONERROR, 1L);

			::curl_easy_setopt(curl, CURLOPT_CAINFO, nullptr);
			::curl_easy_setopt(curl, CURLOPT_CAPATH, _root_ca_certificates);
			::curl_easy_setopt(curl, CURLOPT_SSLCERT, _device_certificate);
			::curl_easy_setopt(curl, CURLOPT_SSLKEY, _device_private_key);

			return curl;
		}

		static HTTP::Status_code _curl_send_request(CURL *curl)
		{
			auto curl_code { ::curl_easy_perform(curl) };
			if (curl_code != CURLE_OK)
				Genode::error("curl failed: ", ::curl_easy_strerror(curl_code));

			long response_code;
			::curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, &response_code);

			::curl_easy_cleanup(curl);

			/* `response_code` is zero at this point if not response has been
			 * received */
			return response_code == 0 ?
				static_cast<HTTP::Status_code>(HTTP::STATUS_INTERNAL_SERVER_ERROR) :
				static_cast<HTTP::Status_code>(response_code);
		}

		/* callback that reads data from a file */
		static Genode::size_t _curl_read_file(char *buf,
		                                      Genode::size_t,
		                                      Genode::size_t nmemb,
		                                      void *userdata)
		{
			auto file { static_cast<Readonly_file *>(userdata) };

			try {
				return file->read(buf, nmemb);

			} catch (Readonly_file::Read_exception const &) {
				Genode::error("failed to read file for transfer");
				return CURL_READFUNC_ABORT;
			}
		}

		/* callback that parses and stores HTTP response JSON payloads */
		static Genode::size_t _curl_parse_json_response(void *ptr,
		                                                Genode::size_t size,
		                                                Genode::size_t nmemb,
		                                                void *userdata)
		{
			auto data { static_cast<char const *>(ptr) };

			auto self { static_cast<File_uploader *>(userdata) };

			try {
				self->_json_response = self->_json_parser.parse(data, nmemb * size);

			} catch (JSON::Exception const &) {
				Genode::error("received invalid JSON response from Azure while uploading file!");
			}

			return nmemb * size;
		}

		/* callback that ignores HTTP responses */
		static Genode::size_t _curl_ignore_response(void *,
		                                            Genode::size_t size,
		                                            Genode::size_t nmemb,
		                                            void *)
		{
			return nmemb * size;
		}

	public:

		explicit File_uploader(Azure::Url file_upload_endpoint,
		                       Azure::Url file_upload_notify_endpoint,
		                       char const *root_ca_certificates,
		                       char const *device_certificate,
		                       char const *device_private_key)
		:
			_file_upload_endpoint { file_upload_endpoint },
			_file_upload_notify_endpoint { file_upload_notify_endpoint },
			_root_ca_certificates { root_ca_certificates },
			_device_certificate { device_certificate },
			_device_private_key { device_private_key }
		{ }

		File_uploader(File_uploader const &)    = delete;
		File_uploader &operator=(File_uploader) = delete;
		File_uploader(File_uploader &&)         = delete;
		void operator=(File_uploader &&)        = delete;

		/* upload the file with path `file_path` to the blob storage associated
		 * with our Azure IoT Hub, see https://learn.microsoft.com/en-us/azure/iot-hub/iot-hub-devguide-file-upload
		 * for additional information */
		HTTP::Status_code upload_file(Path const &file_path)
		{
			/* get blob storage filename */
			auto cloud_file_name { _get_cloud_file_name(file_path) };

			/* get SAS URI parameters */
			auto get_sas_uri_parameter_response_code {
				_get_sas_uri_parameters(cloud_file_name) };

			if (get_sas_uri_parameter_response_code != HTTP::STATUS_OK)
				return get_sas_uri_parameter_response_code;

			auto const &sas_uri_parameters { _json_response };

			/* construct SAS URI */
			Azure::Url sas_uri;

			auto get_sas_uri_response_code {
				_get_sas_uri(sas_uri_parameters, sas_uri) };

			if (get_sas_uri_response_code != HTTP::STATUS_OK)
				return get_sas_uri_response_code;

			/* upload file */
			auto upload_response_code { _upload_file(file_path, sas_uri) };

			/* notify IoT Hub */
			return _notify_iot_hub(sas_uri_parameters, upload_response_code);
		}

		/* upload each file in a given directory */
		HTTP::Status_code upload_directory(Path const &directory_path)
		{
			Genode::log("uploading every file in directory '",
			            directory_path, "'");

			try {
				Directory directory { directory_path };

				HTTP::Status_code response_code { HTTP::STATUS_OK };

				directory.for_each_file([this, &response_code](Path const &file_path) {
					if (HTTP::okay(response_code))
						response_code = upload_file(file_path);
				});

				return response_code;

			} catch (Directory::Open_exception const &e) {
				Genode::error(
					"failed to open directory '", directory_path, "': ",
					Genode::Cstring { ::strerror(e.what_errno) });
				return HTTP::STATUS_INTERNAL_SERVER_ERROR;

			} catch (Directory::Iterate_exception const &e) {
				Genode::error(
					"failed to iterate directory '", directory_path, "': ",
					Genode::Cstring { ::strerror(e.what_errno) });
				return HTTP::STATUS_INTERNAL_SERVER_ERROR;
			}
		}
};
