/*
 * \brief  Azure SDK IoT Provisioning client
 * \author Timo Nicolai
 * \date   2023-03-06
 */

/*
 * Copyright (C) 2023 gapfruit AG
 */

#pragma once

/* Azure includes */
extern "C" {

#include <azure/az_iot.h>

}

/* local includes */
#include <azure_iot_client/azure/check.h>
#include <azure_iot_client/azure/types.h>
#include <azure_iot_client/azure/util.h>
#include <azure_iot_client/iot_client.h>
#include <azure_iot_client/json/json.h>
#include <azure_iot_client/mqtt/types.h>

namespace Azure {
	class IoT_provisioning_client;
}


class Azure::IoT_provisioning_client
	: protected IoT_client<::az_iot_provisioning_client>
{
	public:

		using Register_response = ::az_iot_provisioning_client_register_response;

	private:

		static constexpr char const *GLOBAL_HOSTNAME =
			"global.azure-devices-provisioning.net";

		Id_scope        _id_scope;
		Registration_id _registration_id;

		MQTT::Topic  _register_response_topic   { };
		JSON::Object _register_response_payload { };

		Genode::size_t _mqtt_register_topic_request_id { 0 };

	public:

		explicit IoT_provisioning_client(Id_scope const &id_scope,
		                                 Registration_id const &registration_id)
		:
			_id_scope { id_scope },
			_registration_id { registration_id }
		{
			check(::az_iot_provisioning_client_init(&_client,
			                                        to_span(GLOBAL_HOSTNAME),
			                                        to_span(_id_scope),
			                                        to_span(_registration_id),
			                                        nullptr),
			      "initialize IoT Provisioning client");
		}

		MQTT::Client_id mqtt_client_id() const
		{
			return _get<MQTT::Client_id>("get IoT Provisioning MQTT client id",
			                             ::az_iot_provisioning_client_get_client_id);
		}

		MQTT::Hostname mqtt_hostname() const
		{
			return GLOBAL_HOSTNAME;
		}

		MQTT::Username mqtt_username() const
		{
			return _get<MQTT::Username>("get IoT Provisioning MQTT username",
			                            ::az_iot_provisioning_client_get_user_name);
		}

		MQTT::Topic mqtt_register_topic()
		{
			return _get<MQTT::Topic>(
				"get register topic",
				::az_iot_provisioning_client_register_get_publish_topic);
		}

		MQTT::Topic mqtt_query_status_topic(Register_response const &response)
		{
			return _get<MQTT::Topic>(
				"get query status topic",
				::az_iot_provisioning_client_query_status_get_publish_topic,
				response.operation_id);
		}

		Register_response mqtt_parse_register_response(
			MQTT::Topic const &topic,
			JSON::Object const &payload)
		{
			_register_response_topic = topic;
			_register_response_payload = payload;

			::az_iot_provisioning_client_register_response response;

			check(::az_iot_provisioning_client_parse_received_topic_and_payload(
				&_client,
				to_span(_register_response_topic),
				to_span(_register_response_payload.repr()),
				&response),
			      "parse register response");

			return response;
		}

		static bool mqtt_register_response_complete(
			Register_response const &response)
		{
			return ::az_iot_provisioning_client_operation_complete(
				response.operation_status);
		}
};
