/*
 * \brief  JSON utilities
 * \author Timo Nicolai
 * \date   2023-03-01
 */

/*
 * Copyright (C) 2023 gapfruit AG
 */

#pragma once

/* Genode includes */
#include <base/exception.h>
#include <base/log.h>
#include <util/string.h>

/* Json-C includes */
extern "C" {

#include <json-c/json.h>

}

namespace Genode {
	class Expanding_reporter;
	class Xml_node;
}

namespace JSON {
	struct Exception;
	class Object;
	class Parser;
}


struct JSON::Exception : Genode::Exception { };


class JSON::Object
{
	friend class Parser;

	private:

		::json_object *_object;
		bool           _owned;
		::json_type    _type { };

		explicit Object(::json_object *object, bool owned)
		:
			_object { object },
			_owned { owned }
		{
			_type = ::json_object_get_type(_object);
		}

		static ::json_object *_check_alloc(::json_object *object)
		{
			if (!object) {
				Genode::error("JSON: out of memory");
				throw Exception { };
			}

			return object;
		}

		void _check_type(::json_type type) const
		{
			if (_type != type) {
				Genode::error("JSON: not an instance of '", ::json_type_to_name(type), "'");
				throw Exception { };
			}
		}

	public:

		explicit Object()
		:
			_object { _check_alloc(::json_object_new_object()) },
			_owned { true }
		{
			_type = ::json_object_get_type(_object);
		}

		~Object()
		{
			if (_owned)
				::json_object_put(_object);
		}

		Object(Object const &other)
		:
			_object { other._owned ? ::json_object_get(other._object) : other._object },
			_owned { other._owned },
			_type { other._type }
		{ }

		Object &operator=(Object const &other)
		{
			if (_owned)
				::json_object_put(_object);

			_object = other._owned ? ::json_object_get(other._object) : other._object;
			_owned = other._owned;
			_type = other._type;

			return *this;
		}

		bool operator==(Object const &other) const
		{
			return ::json_object_equal(_object, other._object);
		}

		bool operator!=(Object const &other) const
		{
			return !operator==(other);
		}

		Object operator[](char const *key) const
		{
			auto sub_object { ::json_object_object_get(_object, key) };
			if (!sub_object) {
				Genode::error("JSON: no such key: '", key, "'");
				throw Exception { };
			}

			return Object { sub_object, false };
		}

		static Object from_string(char const *str, Genode::size_t len)
		{
			return Object { _check_alloc(::json_object_new_string_len(str, static_cast<int>(len))), true };
		}

		static Object from_string(char const *str)
		{
			return Object { _check_alloc(::json_object_new_string(str)), true };
		}

		template<Genode::size_t CAPACITY>
		static Object from_string(Genode::String<CAPACITY> const &str)
		{
			return from_string(str.string());
		}

		static Object from_int(Genode::int64_t i)
		{
			return Object { _check_alloc(::json_object_new_int64(i)), true };
		}

		static Object from_bool(bool b)
		{
			return Object { _check_alloc(::json_object_new_boolean(b)), true };
		}

		class List_builder
		{
			private:

				::json_object *_list;

			public:

				explicit List_builder(::json_object *list) : _list { list } { }

				List_builder const &add(Object const &object) const
				{
					/* TODO: do we need to put array elements? */
					::json_object_array_add(_list, ::json_object_get(object._object));
					return *this;
				}

				bool empty() const
				{
					return ::json_object_array_length(_list) == 0;
				}

				Object build() const
				{
					return Object { _list, true };
				}
		};

		static List_builder make_list()
		{
			return List_builder { _check_alloc(::json_object_new_array()) };
		}

		/* return an owned copy of this object, this is useful e.g. for taking
		 * shared ownership of a sub-object returned via operator[]
		 */
		JSON::Object copy() const
		{ return JSON::Object { ::json_object_get(_object), true }; }

		bool is_object() const
		{
			return _type == ::json_type_object;
		}

		/* returns `true` if and only if this is an object which contains no
		 * keys */
		bool is_empty_object() const
		{
			if (!is_object())
				return false;

			bool empty { true };

			for_each_key([&empty](char const *) {
				empty = false;
			});

			return empty;
		}

		bool has_key(char const *key) const
		{
			return ::json_object_object_get(_object, key);
		}

		template<Genode::size_t CAPACITY>
		bool has_key(Genode::String<CAPACITY> const &key) const
		{
			return has_key(key.string());
		}

		template<typename FUNC>
		void for_each_key(FUNC &&fn) const
		{
			json_object_object_foreach(_object, key, val) {
				fn(key);
			}
		}

		void add_key(char const *key, Object const &sub_object)
		{
			auto sub_owned { ::json_object_get(sub_object._object) };

			::json_object_object_add(_object, key, sub_owned);
		}

		template<Genode::size_t CAPACITY>
		void add_key(Genode::String<CAPACITY> const &key,
		             Object const &sub_object)
		{
			add_key(key.string(), sub_object);
		}

		void remove_key(char const *key)
		{
			::json_object_object_del(_object, key);
		}

		template<Genode::size_t CAPACITY>
		void remove_key(Genode::String<CAPACITY> const &key)
		{
			remove_key(key.string());
		}

		bool is_list() const
		{
			return _type == ::json_type_array;
		}

		Genode::size_t list_length() const
		{
			_check_type(::json_type_array);

			return ::json_object_array_length(_object);
		}

		template<typename FUNC>
		void for_each_in_list(FUNC &&fn) const
		{
			_check_type(::json_type_array);

			for (Genode::size_t i { 0 }; i < list_length(); ++i) {
				Object sub_object { ::json_object_array_get_idx(_object, i), false };
				fn(sub_object);
			}
		}

		bool is_string() const
		{
			return _type == ::json_type_string;
		}

		template<typename T>
		T as_string() const
		{
			_check_type(::json_type_string);

			return ::json_object_get_string(_object);
		}

		/* get JSON object represented as a string, in contrast to `as_string`
		 * this works for any object */
		char const *repr() const
		{
			return ::json_object_get_string(_object);
		}

		/* tranform JSON to XML */
		void xml(Genode::Expanding_reporter &reporter) const;
};


class JSON::Parser
{
	private:

		::json_tokener *_tokener { nullptr };

	public:

		explicit Parser()
		{
			_tokener = ::json_tokener_new();
			if (!_tokener) {
				Genode::error("JSON: failed to create tokenizer");
				throw Exception { };
			}
		}

		~Parser()
		{
			if (_tokener)
				::json_tokener_free(_tokener);
		}

		Parser(Parser const &)            = delete;
		Parser &operator=(Parser const &) = delete;

		Object parse(char const *str, Genode::size_t len) const
		{
			auto json { json_tokener_parse_ex(_tokener, str, static_cast<int>(len)) };
			if (!json) {
				Genode::error("JSON: failed to parse '", str, "'");
				throw Exception { };
			}

			return Object { json, true };
		}

		Object parse(char const *str) const
		{
			return parse(str, Genode::strlen(str));
		}

		template<Genode::size_t CAPACITY>
		Object parse(Genode::String<CAPACITY> const &str) const
		{
			return parse(str.string());
		}

		/* Create an equivalent JSON object from an XML object.
		 *
		 * Some examples will help to illustrate this:
		 *
		 * input: <e/>
		 * output: {
		 *           "@type": "e"
		 *         }
		 *
		 * input: <e key="value">hello world</e>
		 * output: {
		 *           "@type": "e",
		 *           "key": "value",
		 *           "@content": "hello world"
		 *         }
		 *
		 * input: <e> <a/> <b/> <a/> </e>
		 * output: {
		 *           "@type": "e",
		 *           "@content": [{ "@type": "a" }, { "@type": "b" }, { "@type": "a" }]
		 *         }
		 */
		static Object parse(Genode::Xml_node const &xml);
};
