/*
 * \brief  MQTT client
 * \author Timo Nicolai
 * \date   2023-02-23
 */

/*
 * Copyright (C) 2023 gapfruit AG
 */

#pragma once

/* Genode includes */
#include <base/signal.h>
#include <timer_session/connection.h>

/* Azure includes */
extern "C" {

#include <azure/az_iot.h>

}

/* Mosquitto includes */
extern "C" {

#include <mosquitto.h>

}

/* local includes */
#include <azure_iot_client/json/json.h>
#include <azure_iot_client/mqtt/check.h>
#include <azure_iot_client/mqtt/types.h>

namespace Genode {
	class Env;
}

namespace MQTT {
	class Client;
}


class MQTT::Client
{
	public:

		using Callback = void (*)(Topic const &, JSON::Object const &);

	private:

		enum : int
		{
			DEFAULT_PORT         = AZ_IOT_DEFAULT_MQTT_CONNECT_PORT,
			DEFAULT_KEEPALIVE    = AZ_IOT_DEFAULT_MQTT_CONNECT_KEEPALIVE_SECONDS,
			SSL_VERIFY_PEER      = 1,
			RECONNECT_TIMEOUT_MS = 10'000,
		};

		enum { NUM_SUBSCRIPTIONS_MAX = 10 };

		struct Subscription
		{
			Topic    topic    { };
			Callback callback { };
		};

		Hostname   _server_hostname                    { };
		Username   _server_username                    { };
		void     (*_server_on_connect)(Genode::size_t) { nullptr };

		static inline Genode::size_t  _num_clients            { 0 };
		::mosquitto                  *_client                 { nullptr };
		int                           _client_loop_last_error { MOSQ_ERR_SUCCESS };
		bool                          _client_connected       { false };

		::Timer::Connection            _reconnect_timer;
		Genode::Signal_handler<Client> _reconnect_timer_handler;
		bool                           _reconnect_timeout_active { false };
		Genode::size_t                 _reconnect_counter        { 0 };

		Genode::size_t _num_subscriptions { 0 };
		Subscription   _subscriptions[NUM_SUBSCRIPTIONS_MAX];

		JSON::Parser _json_parser { };

		static void _on_log(::mosquitto *, void *, int level, char const *message);

		static void _on_connect(::mosquitto *, void *userdata, int rc);

		static void _on_message(::mosquitto *,
		                        void *userdata,
		                        ::mosquitto_message const *message);

		void _configure_mqtt() const;

		void _configure_tls(char const *root_ca_certificates,
		                    char const *device_certificate,
		                    char const *device_private_key) const;

		void _add_subscription(Topic const &topic, Callback callback);

		void _reconnect();

		void _reconnect_on_error(int rc, char const *action)
		{
			/* To avoid reconnecting to many times in quick succession, we
			 * set this flag indicating that we've recently retried and only
			 * retry again when it is reset by a timer. */
			if (_reconnect_timeout_active)
				return;

			try {
				check(rc, action);

			} catch (MQTT::Exception const &e) {
				if (e.rc == MOSQ_ERR_NO_CONN || e.rc == MOSQ_ERR_CONN_LOST) {
					_client_connected = false;

					_reconnect();

				} else {
					throw;
				}
			}
		}

		void _handle_reconnect_timer()
		{
			_reconnect_timeout_active = false;
		}

	public:

		explicit Client(Genode::Env &env,
		                Client_id const &client_id,
		                char const *root_ca_certificates,
		                char const *device_certificate,
		                char const *device_private_key);
		~Client();

		Client(Client const &)    = delete;
		Client &operator=(Client) = delete;

		void run();

		void connect(Hostname const &hostname,
		             Username const &username,
		             void (*on_connect)(Genode::size_t));

		void publish(Topic const &topic);
		void publish(Topic const &topic, JSON::Object const &payload_json);

		void subscribe(Topic const &topic, Callback callback);
};
