/*
 * \brief  Filesystem utilities
 * \author Timo Nicolai
 * \date   2023-06-03
 */

/*
 *  Copyright (C) 2023 Gapfruit AG
 */

#pragma once

/* Genode includes */
#include <base/log.h>
#include <os/vfs.h>
#include <util/string.h>

/* libc includes */
#include <dirent.h>
#include <errno.h>
#include <libgen.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>

namespace Azure_iot_client
{
	using Path = Genode::Directory::Path;

	class Directory;
	class Readonly_file;
}


class Azure_iot_client::Directory
{
	private:

		Path  _directory_path;
		DIR  *_directory;

	public:

		struct Exception : Genode::Exception
		{
			int what_errno;

			explicit Exception(int what_errno) : what_errno { what_errno } { };
		};

		struct Open_exception : Exception
		{
			using Exception::Exception;
		};

		struct Iterate_exception : Exception
		{
			using Exception::Exception;
		};

		explicit Directory(Path directory_path)
		:
			_directory_path { directory_path },
			_directory { ::opendir(directory_path.string()) }
		{
			if (!_directory)
				throw Open_exception { errno };
		}

		~Directory()
		{
			if (_directory)
				::closedir(_directory);
		}

		Directory(Directory const &)    = delete;
		Directory &operator=(Directory) = delete;
		Directory(Directory &&)         = delete;
		void operator=(Directory &&)    = delete;

		template<typename FUNC>
		void for_each_file(FUNC &&func)
		{
			errno = 0;

			::dirent *entry;
			while ((entry = ::readdir(_directory))) {
				if (entry->d_type != DT_REG)
					continue;

				Path file_path { _directory_path,
				                 "/",
				                 Genode::Cstring { entry->d_name } };

				func(file_path);
			}

			if (errno != 0)
				throw Iterate_exception { errno };

			::rewinddir(_directory);
		}
};


class Azure_iot_client::Readonly_file
{
	private:

		FILE *_file;

	public:

		struct Exception : Genode::Exception { };
		struct Open_exception : Exception { };
		struct Read_exception : Exception { };

		explicit Readonly_file(Path const &file_path)
		:
			_file { ::fopen(file_path.string(), "rb") }
		{
			if (!_file)
				throw Open_exception { };
		}

		~Readonly_file()
		{
			if (_file)
				::fclose(_file);
		}

		Readonly_file(Readonly_file const &)    = delete;
		Readonly_file &operator=(Readonly_file) = delete;
		Readonly_file(Readonly_file &&)         = delete;
		void operator=(Readonly_file &&)        = delete;

		long size()
		{
			if (::fseek(_file, 0L, SEEK_END) != 0)
				throw Read_exception { };

			long size;
			if ((size = ::ftell(_file)) == -1L)
				throw Read_exception { };

			::rewind(_file);

			return size;
		}

		Genode::size_t read(char *buf, Genode::size_t buf_size) const
		{
			Genode::size_t read_size;
			if ((read_size = ::fread(buf, 1, buf_size, _file)) != buf_size) {
				if (!::feof(_file))
					throw Read_exception { };
			}

			return read_size;
		}
};
